"""A simple configuration parsesr."""

class ConfigFormatError(Exception):
    """Configuration file format error."""
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class Config:
    """A class for configuration of the neural net."""
    
    # section names
    SEC_DATA       = 'data'
    SEC_PARAMETERS = 'parameters'

    # attribute names
    DATA_TRAIN_FILE       = 'train_data_file'
    DATA_TRAIN_INDEX_FILE = 'train_index_file'
    DATA_UNLABELED_FILE   = 'unlabeled_data_file'
    DATA_VAL_FILE         = 'val_data_file'
    DATA_TEST_FILE        = 'test_data_file'
    DATA_OUTPUT_DIR       = 'output_dir'
    DATA_INIT_MODEL_FILE  = 'init_model_file'

    PAR_LEARN_RATE       = 'learn_rate'
    PAR_INIT_SCALE       = 'init_scale'
    PAR_INIT_MOMENTUM    = 'init_momentum'
    PAR_SWITCH_EPOCH     = 'switch_epoch'
    PAR_FINAL_MOMENTUM   = 'final_momentum'
    PAR_WEIGHT_DECAY     = 'weight_decay'
    PAR_MINIBATCH_SIZE   = 'minibatch_size'
    PAR_NUM_MINIBATCHES  = 'num_minibatches'
    PAR_NUM_EPOCHS       = 'num_epochs'
    PAR_EPOCH_TO_DISPLAY = 'epoch_to_display'
    PAR_DISPLAY_WINC     = 'display_winc'
    PAR_EPOCH_TO_SAVE    = 'epoch_to_save'
    PAR_SHOW_ACCURACY    = 'show_accuracy'
    PAR_RANDOM_SEED      = 'random_seed'

    PAR_C               = 'C'
    PAR_C_STAR          = 'C_star'
    PAR_SIM_WEIGHT      = 'sim_weight'
    PAR_GRAPH_CUT_ITER  = 'graph_cut_iter'
    PAR_SIM_THRESHOLD   = 'sim_threshold'
    PAR_TOP_K           = 'top_k'
    PAR_UNIFORM_WEIGHT  = 'uniform_weight'
    PAR_UNLABELED_COUNT = 'unlabeled_count'
    PAR_FOREGROUND_BIAS = 'foreground_bias'
    PAR_FREEZE_EPOCHS   = 'freeze_epochs'
    PAR_U_LOSS_AUGMENTED = 'u_loss_augmented'
    PAR_INCONSISTENT_TO_FILTER = 'inconsistent_to_filter'
    PAR_VERBOSE         = 'verbose'
    PAR_USE_CARDINALITY = 'use_cardinality'
    PAR_CARDINALITY_SENSITIVITY = 'cardinality_sensitivity'
    PAR_DD_ITER         = 'dd_iter'
    PAR_DD_LEARN_RATE   = 'dd_learn_rate'
    PAR_DD_UNARY_SPLIT  = 'dd_unary_split'
    PAR_USE_IMAGE_DEPENDENT_UNARY = 'use_image_dependent_unary'
    PAR_SELF_TRAINING   = 'self_training'

    PAR_INTELLIGENT_STOPPING = 'intelligent_stopping'

    def __init__(self, file_name):
        """Set default option values. Read configurations from a text file."""
        # data files
        self.train_data_file = ''
        self.train_index_file = ''
        self.unlabeled_data_file = ''
        self.extra_unlabeled = False
        self.val_data_file = ''
        self.test_data_file = ''
        self.init_model_file = ''
        self.is_val = False
        self.is_test = False
        self.output_dir = ''
        self.output_filename_pattern = 'm%d'
        self.is_output = False

        # parameters for training
        self.learn_rate = 0.001
        self.init_scale = 0.001
        self.init_momentum = 0.5
        self.switch_epoch = 20
        self.final_momentum = 0.9
        self.weight_decay = 0
        self.minibatch_size = 100
        self.num_minibatches = 10
        self.epoch_to_display = 10
        self.display_winc = True 
        self.epoch_to_save = 100
        self.show_accuracy = False
        self.num_epochs = 1000
        self.random_seed = None

        self.C = 1
        self.C_star = 0
        self.sim_weight = 0
        self.graph_cut_iter = 1
        self.sim_threshold = 0.8
        self.top_k = -1
        self.uniform_weight = False
        self.unlabeled_count = 1000
        self.foreground_bias = 0.0
        self.freeze_epochs = 1
        self.u_loss_augmented = False
        self.inconsistent_to_filter = 0
        self.verbose = 0
        self.use_cardinality = False
        self.cardinality_sensitivity = 5
        self.dd_iter = 10
        self.dd_learn_rate = 1e-1
        self.dd_unary_split = 0.8
        self.use_image_dependent_unary = False
        self.self_training = False

        self.intelligent_stopping = False

        self._parse_cfg_file(file_name)

    def _parse_cfg_file(self, file_name):
        """Parse a configuration file."""
        import ConfigParser

        cfg = ConfigParser.ConfigParser()
        cfg.read(file_name)
        # make sure the config file contains all required sections
        self._check_cfg_sections(cfg)
        
        # parse data section
        if cfg.has_option(Config.SEC_DATA, Config.DATA_TRAIN_FILE):
            self.train_data_file = cfg.get(Config.SEC_DATA, Config.DATA_TRAIN_FILE)
        else:
            raise ConfigFormatError('No training data.')
        if cfg.has_option(Config.SEC_DATA, Config.DATA_TRAIN_INDEX_FILE):
            self.train_index_file = cfg.get(Config.SEC_DATA, Config.DATA_TRAIN_INDEX_FILE)
        else:
            raise ConfigFormatError('No index file.')
        if cfg.has_option(Config.SEC_DATA, Config.DATA_UNLABELED_FILE):
            self.unlabeled_data_file = cfg.get(Config.SEC_DATA, Config.DATA_UNLABELED_FILE)
            self.extra_unlabeled = True

        if cfg.has_option(Config.SEC_DATA, Config.DATA_VAL_FILE):
            self.val_data_file = cfg.get(Config.SEC_DATA, Config.DATA_VAL_FILE)
            self.is_val = True

        if cfg.has_option(Config.SEC_DATA, Config.DATA_TEST_FILE):
            self.test_data_file = cfg.get(Config.SEC_DATA, Config.DATA_TEST_FILE)
            self.is_test = True

        if cfg.has_option(Config.SEC_DATA, Config.DATA_INIT_MODEL_FILE):
            self.init_model_file = cfg.get(Config.SEC_DATA, Config.DATA_INIT_MODEL_FILE)
            
        if cfg.has_option(Config.SEC_DATA, Config.DATA_OUTPUT_DIR):
            self.output_dir = cfg.get(Config.SEC_DATA, Config.DATA_OUTPUT_DIR)
            self.is_output = True

        # parse parameter section
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_SIM_THRESHOLD):
            self.sim_threshold = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_SIM_THRESHOLD))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_LEARN_RATE):
            self.learn_rate = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_LEARN_RATE))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_INIT_SCALE):
            self.init_scale = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_INIT_SCALE))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_INIT_MOMENTUM):
            self.init_momentum = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_INIT_MOMENTUM))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_SWITCH_EPOCH):
            self.switch_epoch = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_SWITCH_EPOCH))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_FINAL_MOMENTUM):
            self.final_momentum = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_FINAL_MOMENTUM))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_WEIGHT_DECAY):
            self.weight_decay = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_WEIGHT_DECAY))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_MINIBATCH_SIZE):
            self.minibatch_size = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_MINIBATCH_SIZE))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_NUM_MINIBATCHES):
            self.num_minibatches = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_NUM_MINIBATCHES))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_NUM_EPOCHS):
            self.num_epochs = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_NUM_EPOCHS))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_EPOCH_TO_DISPLAY):
            self.epoch_to_display = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_EPOCH_TO_DISPLAY))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_DISPLAY_WINC):
            self.display_winc = (cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_DISPLAY_WINC).lower() == 'true')
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_SHOW_ACCURACY):
            self.show_accuracy = (cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_SHOW_ACCURACY).lower() == 'true')
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_EPOCH_TO_SAVE):
            self.epoch_to_save = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_EPOCH_TO_SAVE))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_RANDOM_SEED):
            self.random_seed = int(cfg.get(Config.SEC_PARAMETERS,
                Config.PAR_RANDOM_SEED))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_C):
            self.C = float(cfg.get(Config.SEC_PARAMETERS, Config.PAR_C))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_C_STAR):
            self.C_star = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_C_STAR))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_SIM_WEIGHT):
            self.sim_weight = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_SIM_WEIGHT))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_GRAPH_CUT_ITER):
            self.graph_cut_iter = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_GRAPH_CUT_ITER))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_TOP_K):
            self.top_k = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_TOP_K))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_UNIFORM_WEIGHT):
            self.uniform_weight = (cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_UNIFORM_WEIGHT).lower() == 'true')
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_UNLABELED_COUNT) \
                and self.extra_unlabeled:
            self.unlabeled_count = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_UNLABELED_COUNT))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_FOREGROUND_BIAS):
            self.foreground_bias = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_FOREGROUND_BIAS))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_FREEZE_EPOCHS):
            self.freeze_epochs = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_FREEZE_EPOCHS))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_U_LOSS_AUGMENTED):
            self.u_loss_augmented = (cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_U_LOSS_AUGMENTED).lower() == 'true')
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_INCONSISTENT_TO_FILTER):
            self.inconsistent_to_filter = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_INCONSISTENT_TO_FILTER))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_VERBOSE):
            self.verbose = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_VERBOSE))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_USE_CARDINALITY):
            self.use_cardinality = (cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_USE_CARDINALITY).lower() == 'true')
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_CARDINALITY_SENSITIVITY):
            self.cardinality_sensitivity = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_CARDINALITY_SENSITIVITY))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_DD_ITER):
            self.dd_iter = int(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_DD_ITER))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_DD_LEARN_RATE):
            self.dd_learn_rate = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_DD_LEARN_RATE))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_DD_UNARY_SPLIT):
            self.dd_unary_split = float(cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_DD_UNARY_SPLIT))
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_USE_IMAGE_DEPENDENT_UNARY):
            self.use_image_dependent_unary = (cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_USE_IMAGE_DEPENDENT_UNARY).lower() == 'true')
        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_SELF_TRAINING):
            self.self_training = (cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_SELF_TRAINING).lower() == 'true')

        if cfg.has_option(Config.SEC_PARAMETERS, Config.PAR_INTELLIGENT_STOPPING):
            self.intelligent_stopping = (cfg.get(Config.SEC_PARAMETERS, 
                Config.PAR_INTELLIGENT_STOPPING).lower() == 'true')

    def _check_cfg_sections(self, cfg):
        """The config parser cfg should already have the config file loaded.
        """
        secs = cfg.sections()

        n_data_sec = 0
        n_par_sec = 0
        current_layer = 0

        for section in secs:
            if section == Config.SEC_DATA:
                n_data_sec += 1
                if n_data_sec > 1:
                    raise ConfigFormatError('Multiple data sections.')
            elif section == Config.SEC_PARAMETERS:
                n_par_sec += 1
                if n_par_sec > 1:
                    raise ConfigFormatError('Multiple parameter sections.')
            else:
                raise ConfigFormatError('Unknown section name [%s]' % section)

        return True

    def display(self):
        """For debug use only."""
        self.display_data_info()
        self.display_parameters()

    def display_data_info(self):
        """data section"""
        print '[' + Config.SEC_DATA + ']'
        print Config.DATA_TRAIN_FILE +  '=' + self.train_data_file
        print Config.DATA_TRAIN_INDEX_FILE + '=' + self.train_index_file
        print Config.DATA_UNLABELED_FILE + '=' + self.unlabeled_data_file
        print 'extra_unlabeled=' + str(self.extra_unlabeled)
        print Config.DATA_VAL_FILE +    '=' + self.val_data_file
        print 'is_val=' + str(self.is_val)
        print Config.DATA_TEST_FILE +   '=' + self.test_data_file
        print 'is_test=' + str(self.is_test)
        print Config.DATA_INIT_MODEL_FILE + '=' + self.init_model_file
        print Config.DATA_OUTPUT_DIR +  '=' + self.output_dir
        print 'is_output=' + str(self.is_output)
        print ''

    def display_parameters(self):
        """parameter section"""
        print self

    def __str__(self):
        s = ''
        s += '[' + Config.SEC_PARAMETERS + ']\n'
        s += Config.PAR_LEARN_RATE +       '=' + str(self.learn_rate) + '\n'
        s += Config.PAR_INIT_SCALE +       '=' + str(self.init_scale) + '\n'
        s += Config.PAR_INIT_MOMENTUM +    '=' + str(self.init_momentum) + '\n'
        s += Config.PAR_SWITCH_EPOCH +     '=' + str(self.switch_epoch) + '\n'
        s += Config.PAR_FINAL_MOMENTUM +   '=' + str(self.final_momentum) + '\n'
        s += Config.PAR_WEIGHT_DECAY +     '=' + str(self.weight_decay) + '\n'
        s += Config.PAR_MINIBATCH_SIZE +   '=' + str(self.minibatch_size) + '\n'
        s += Config.PAR_NUM_MINIBATCHES +  '=' + str(self.num_minibatches) + '\n'
        s += Config.PAR_NUM_EPOCHS +       '=' + str(self.num_epochs) + '\n'
        s += Config.PAR_EPOCH_TO_DISPLAY + '=' + str(self.epoch_to_display) + '\n'
        s += Config.PAR_DISPLAY_WINC +     '=' + str(self.display_winc) + '\n'
        s += Config.PAR_EPOCH_TO_SAVE +    '=' + str(self.epoch_to_save) + '\n'
        s += Config.PAR_SIM_THRESHOLD +    '=' + str(self.sim_threshold) + '\n'
        s += Config.PAR_C +                '=' + str(self.C) + '\n'
        s += Config.PAR_C_STAR +           '=' + str(self.C_star) + '\n'
        s += Config.PAR_SIM_WEIGHT +       '=' + str(self.sim_weight) + '\n'
        s += Config.PAR_GRAPH_CUT_ITER +   '=' + str(self.graph_cut_iter) + '\n'
        s += Config.PAR_TOP_K +            '=' + str(self.top_k) + '\n'
        s += Config.PAR_UNIFORM_WEIGHT +   '=' + str(self.uniform_weight) + '\n'
        if self.extra_unlabeled:
            s += Config.PAR_UNLABELED_COUNT + '=' + str(self.unlabeled_count) + '\n'
        s += Config.PAR_FOREGROUND_BIAS +  '=' + str(self.foreground_bias) + '\n'
        s += Config.PAR_FREEZE_EPOCHS +    '=' + str(self.freeze_epochs) + '\n'
        s += Config.PAR_U_LOSS_AUGMENTED + '=' + str(self.u_loss_augmented) + '\n'
        s += Config.PAR_INCONSISTENT_TO_FILTER + '=' + str(self.inconsistent_to_filter) + '\n'
        s += Config.PAR_VERBOSE +          '=' + str(self.verbose) + '\n'
        s += Config.PAR_USE_CARDINALITY +  '=' + str(self.use_cardinality) + '\n'
        s += Config.PAR_CARDINALITY_SENSITIVITY + '=' + str(self.cardinality_sensitivity) + '\n'
        s += Config.PAR_DD_ITER +          '=' + str(self.dd_iter) + '\n'
        s += Config.PAR_DD_LEARN_RATE +    '=' + str(self.dd_learn_rate) + '\n'
        s += Config.PAR_DD_UNARY_SPLIT +   '=' + str(self.dd_unary_split) + '\n'
        s += Config.PAR_USE_IMAGE_DEPENDENT_UNARY + '=' + str(self.use_image_dependent_unary) + '\n'
        s += Config.PAR_SELF_TRAINING +    '=' + str(self.self_training) + '\n'
        s += Config.PAR_INTELLIGENT_STOPPING + '=' + str(self.intelligent_stopping) + '\n'
        s += '\n'

        return s

