"""
This example file shows how to use the code for our paper [1].

[1] Yujia Li and Richard Zemel. High-Order Regularization for Semi-Supervised
    Learning of Structured Output Problems. ICML 2014.

Yujia Li, 10/2014
"""

import cPickle as pickle
import os
import argparse
import numpy as np

import gnn.nn as nn
import gnn.config as cfg
import imgtools.filter as ft
import imgtools.unary as un
import imgtools.pairwise as pw
import crf.general as gen
import crf.learning as learn
import config
import mrseg


############ data path ##################

_data_dir = 'data'
_split_subset_dir = '%s/bird_subsets' % _data_dir
_processed_data_dir = 'processed_data'

_train_raw_file = '%s/bird_train_raw.pdata' % _data_dir
_val_raw_file = '%s/bird_val_raw.pdata' % _data_dir
_test_raw_file = '%s/bird_test_raw.pdata' % _data_dir
_unlabeled_raw_file = '%s/bird_unlabeled_raw.pdata' % _data_dir

_train_file = '%s/bird_train.pdata' % _processed_data_dir
_val_file = '%s/bird_val.pdata' % _processed_data_dir
_test_file = '%s/bird_test.pdata' % _processed_data_dir
_unlabeled_file = '%s/bird_unlabeled.pdata' % _processed_data_dir

########### model path ###############

_model_dir = 'models'
_unary_dir = '%s/unary' % _model_dir
_pairwise_dir = '%s/pairwise' % _model_dir
_joint_dir = '%s/joint' % _model_dir
_selftrain_dir = '%s/selftrain' % _model_dir
_ssl_dir = '%s/ssl' % _model_dir
_sslcard_dir = '%s/sslcard' % _model_dir

########### other path ###############

_cfg_dir = 'cfg'

########### default parameter settings ###############

_default_feature_window_size = 5
_default_split = 0
_default_n_labeled = 40

def generate_pixel_features(window_size=1):
    if not os.path.exists(_processed_data_dir):
        os.makedirs(_processed_data_dir)

    input_files = [_train_raw_file, _val_raw_file, _test_raw_file, _unlabeled_raw_file]
    output_files= [_train_file, _val_file, _test_file, _unlabeled_file]

    for i in range(len(input_files)):
        print '*****************************************'
        print 'Loading from %s' % input_files[i]

        with open('%s' % input_files[i]) as f:
            d = pickle.load(f)
        d['flist'] = ft.filter_response(d['imgs'], window_size=window_size)
        with open('%s' % output_files[i], 'w') as f:
            pickle.dump(d, f, -1)

def load_data_from_file(fname):
    print 'Loading data from %s' % fname
    with open(fname) as f:
        d = pickle.load(f)

    return d['imgs'], d['segs'], d['flist'], d['hogs']

def load_subset(n_labeled, i_split, imgs, segs, flist, hogs):
    index_fname = '%s/%d/train%d.pdata' % (_split_subset_dir, i_split, n_labeled)
    print 'Loading subset indices from %s' % index_fname
    with open(index_fname) as f:
        d = pickle.load(f)
    mask = d['labeled'] > 0

    s_imgs = None
    if imgs is not None:
        s_imgs = imgs[mask]

    s_segs = None
    if segs is not None:
        s_segs = segs[mask]

    s_flist = None
    if flist is not None:
        s_flist = []
        for i in range(mask.size):
            if mask[i]:
                s_flist.append(flist[i])

    s_hogs = None
    if hogs is not None:
        s_hogs = hogs[mask]

    return s_imgs, s_segs, s_flist, s_hogs

def train_unary_potentials(n_labeled, i_split):
    if not os.path.exists(_unary_dir):
        os.makedirs(_unary_dir)

    train_imgs, train_segs, train_flist, _ = load_data_from_file(_train_file)
    val_imgs, val_segs, val_flist, _ = load_data_from_file(_val_file)

    imgs, segs, flist, _ = load_subset(n_labeled, i_split, train_imgs, train_segs, train_flist, None)

    cfg_file_name = '%s/default_unary.cfg' % _cfg_dir 
    cfg_obj = cfg.Config(cfg_file_name)
    n_imgs = len(imgs)
    cfg_factor = int(round(np.sqrt(200.0 / n_imgs)))
    cfg_obj.switch_epoch = cfg_obj.switch_epoch * cfg_factor
    cfg_obj.num_epochs = cfg_obj.num_epochs * cfg_factor
    cfg_obj.epoch_to_display = cfg_obj.epoch_to_display * cfg_factor

    n_per_class_val = 100
    n_per_class = 100

    unary_model, std, avg, acc = un.generate_default_unary_validation(
            flist, segs, val_flist, val_segs, 2, n_per_class, n_per_class_val, cfg_obj=cfg_obj)
    val_u = un.get_unary(val_flist, unary_model, std, avg)
    acc = (gen.unary_predict(val_u) == val_segs).mean()
    print 'Validation acc %.4f' % acc

    with open('%s/u%d.pdata' % (_unary_dir, n_labeled), 'w') as f:
        pickle.dump({'unary_model' : unary_model, 'std' : std, 'avg' : avg}, f, -1)
        print 'Model saved to %s/u%d.pdata' % (_unary_dir, n_labeled)

def tune_pairwise_potentials(n_labeled, i_split):
    unary_file_name = '%s/u%d.pdata' % (_unary_dir, n_labeled)
    pw_file_name = '%s/pw%d.pdata' % (_pairwise_dir, n_labeled)

    if not os.path.exists(_pairwise_dir):
        os.makedirs(_pairwise_dir)

    val_imgs, val_segs, val_flist, _ = load_data_from_file(_val_file)

    with open(unary_file_name) as f:
        d = pickle.load(f)

    net = nn.NN()
    net.build_net_from_copy(d['unary_model'])
    std = d['std']
    avg = d['avg']

    val_u = un.get_unary(val_flist, net, std, avg)

    print 'Unary accuracy %.4f' % (gen.unary_predict(val_u) == val_segs).mean()

    val_pwlist = [pw.get_uniform_smoothness_pw(val_imgs)]
    sigma, weights, acc = learn.grid_search_pairwise_weights(
            val_imgs, val_u, val_pwlist,
            [np.arange(0.0, 2, 0.2), np.arange(0,10,1), np.arange(0.01,0.1,0.02)], val_segs)

    with open(pw_file_name, 'w') as f:
        pickle.dump({'unary_model' : net.make_copy(), 'std' : std, 'avg' : avg,
            'sigma' : sigma, 'pairwise_weights' : weights}, f, -1)

def jointly_learn_unaries(n_labeled, i_split):
    mr = mrseg.ManifoldRegularizationSeg()
    seg_cfg = config.Config('%s/joint.cfg' % _cfg_dir)
    seg_cfg.train_data_file = _train_file
    seg_cfg.val_data_file = _val_file
    seg_cfg.test_data_file = _test_file
    seg_cfg.unlabeled_data_file = _unlabeled_file
    seg_cfg.train_index_file = '%s/%d/train%d.pdata' % (_split_subset_dir, i_split, n_labeled)
    seg_cfg.init_model_file = '%s/pw%d.pdata' % (_pairwise_dir, n_labeled)

    mr.init_training(seg_cfg)
    mr.train()

    if not os.path.exists(_joint_dir):
        os.makedirs(_joint_dir)

    mr.write_model('%s/joint%d.pdata' % (_joint_dir, n_labeled))

def self_train(n_labeled, i_split):
    mr = mrseg.ManifoldRegularizationSeg()
    seg_cfg = config.Config('%s/selftrain.cfg' % _cfg_dir)

    seg_cfg.train_data_file = _train_file
    seg_cfg.val_data_file = _val_file
    seg_cfg.test_data_file = _test_file
    seg_cfg.unlabeled_data_file = _unlabeled_file
    seg_cfg.train_index_file = '%s/%d/train%d.pdata' % (_split_subset_dir, i_split, n_labeled)
    seg_cfg.init_model_file = '%s/joint%d.pdata' % (_joint_dir, n_labeled)

    #seg_cfg.learn_rate = 1e-8
    #seg_cfg.freeze_epochs = 2
    #seg_cfg.num_epochs = 100

    if not os.path.exists(_selftrain_dir):
        os.makedirs(_selftrain_dir)

    mr.init_training(seg_cfg)
    mr.train()

    mr.write_model('%s/st%d.pdata' % (_selftrain_dir, n_labeled))

def semi_supervised_learning(n_labeled, i_split):
    mr = mrseg.ManifoldRegularizationSeg()
    seg_cfg = config.Config('%s/ssl.cfg' % _cfg_dir)

    seg_cfg.train_data_file = _train_file
    seg_cfg.val_data_file = _val_file
    seg_cfg.test_data_file = _test_file
    seg_cfg.unlabeled_data_file = _unlabeled_file
    seg_cfg.train_index_file = '%s/%d/train%d.pdata' % (_split_subset_dir, i_split, n_labeled)
    seg_cfg.init_model_file = '%s/joint%d.pdata' % (_joint_dir, n_labeled)

    if not os.path.exists(_ssl_dir):
        os.makedirs(_ssl_dir)

    mr.init_training(seg_cfg)
    mr.train()

    mr.write_model('%s/ssl%d.pdata' % (_ssl_dir, n_labeled))

def ssl_cardinality(n_labeled, i_split):
    mr = mrseg.ManifoldRegularizationSeg()
    seg_cfg = config.Config('%s/ssl.cfg' % _cfg_dir)

    seg_cfg.train_data_file = _train_file
    seg_cfg.val_data_file = _val_file
    seg_cfg.test_data_file = _test_file
    seg_cfg.unlabeled_data_file = _unlabeled_file
    seg_cfg.train_index_file = '%s/%d/train%d.pdata' % (_split_subset_dir, i_split, n_labeled)
    seg_cfg.init_model_file = '%s/joint%d.pdata' % (_joint_dir, n_labeled)
 
    seg_cfg.use_cardinality = True

    if not os.path.exists(_sslcard_dir):
        os.makedirs(_sslcard_dir)

    mr.init_training(seg_cfg)
    mr.train()

    mr.write_model('%s/sslcard%d.pdata' % (_sslcard_dir, n_labeled))

def _print_message(msg):
    print ''
    print '< Results obtained here are just for demo purposes, they are not optimized. >'
    print ''
    print '======================================================================================================='
    print msg
    print '======================================================================================================='
    print ''

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--feature', help='Generate pixel features for segmentation.', default=False, action='store_true')
    parser.add_argument('--unary', help='Train unary potentials.', default=False, action='store_true')
    parser.add_argument('--pairwise', help='Tune parameters in pairwise potentials', default=False, action='store_true')
    parser.add_argument('--joint', help='Backprop through the CRF to update unary potentials.', default=False, action='store_true')
    parser.add_argument('--selftrain', help='Run self-training', action='store_true')
    parser.add_argument('--ssl', help='Run our basic semi-supervised learning framework', action='store_true')
    parser.add_argument('--sslcard', help='Run our semi-supervised learning framework with cardinality potentials',
            action='store_true')

    args = parser.parse_args()

    if args.feature:
        _print_message('Generating pixel features for segmentation for each of the training/validation/test/unlabeled datasets.')
        generate_pixel_features(window_size=_default_feature_window_size)
    if args.unary:
        _print_message('Training unary potentials...')
        train_unary_potentials(_default_n_labeled, _default_split)
        _print_message('You can run this multiple times using different random seeds and then choose the best one.')
    if args.pairwise:
        _print_message('Tuning parameters for pairwise potentials...')
        tune_pairwise_potentials(_default_n_labeled, _default_split)
    if args.joint:
        _print_message('Backprop through the CRF to further jointly tune unary potentials.')
        jointly_learn_unaries(_default_n_labeled, _default_split)
    if args.selftrain:
        _print_message('Running self-training baseline.\n' +
                'Learning rate is set to be very small, otherwise there won\'t be any gain using self-training.')
        self_train(_default_n_labeled, _default_split)
    if args.ssl:
        _print_message('Running semi-supervised learning without using high-order cardinality potentials.')
        semi_supervised_learning(_default_n_labeled, _default_split)
    if args.sslcard:
        _print_message('Running SSL with high-order cardinality potentials.')
        ssl_cardinality(_default_n_labeled, _default_split)


