"""
Some small tweaks for mrseg.

Yujia Li, 10/2013
"""

import numpy as np
import cPickle as pickle
import crf.general as gen
import imgtools.unary as un

class ImageDependentUnaryModel(object):
    def make_unary(self, fmat):
        """Generate unary potential for the feature matrix using the current
        model."""
        pass

class TwoGaussiansModel(ImageDependentUnaryModel):
    def __init__(self):
        pass

    def fit_model(self, x1, x2):
        """Fit a mixture of two diagonal Gaussians using given fully labeled
        data. A Gaussian will be fitted to each class and a uniform prior (1/2
        for each class) is assumed."""

        assert x1.shape[1] == x2.shape[1]

        D = x1.shape[1]

        mu = np.empty((2,D), dtype=np.float)
        var = np.empty((2,D), dtype=np.float)

        mu[0] = x1.mean(axis=0)
        mu[1] = x2.mean(axis=0)
        var[0] = x1.var(axis=0, ddof=1)
        var[1] = x2.var(axis=0, ddof=1)

        self.D = D
        self.mu = mu
        self.var = var

    def make_unary(self, fmat):
        """Generate unary potentials for a given feature matrix based on the
        fitted two Gaussians model.

        The unary potentials will be negative log posterior probability for
        each class. Under uniform prior posterior is equal to likelihood.
        """
        n_pix = fmat.shape[0]

        u = np.zeros((n_pix, 2), dtype=np.float32)

        log_sigma = [np.log(self.var[0]).sum(), np.log(self.var[1]).sum()]

        for i in range(2):
            u[:,i] = ((fmat - self.mu[i])**2 / self.var[i]).sum(axis=1) + log_sigma[i]

        return u

def get_image_dependent_unary_for_single_image(imsz, fmat, edge_width, 
        center_block_fraction):
    """
    (imsz, fmat, edge_width, center_block_fraction) --> unary

    This is a simple way to train an image dependent unary model. Here a
    Gaussian is trained to fit the boundary pixels as the background model and 
    another Gaussian is trained to fit the center pixels as the foreground
    model. After training, log posterior probability will be used as unary
    potentials.

    imsz: size of an image, [height, width]
    fmat: feature matrix for this image, size (height*width) * D, where D is
        the number of features extracted for each pixel.
    edge_width: width of the edge to be used to train background model.
    center_block_fraction: if set to r, then a block of (height*r) x (width*r)
        in the center of the image will be used to train the foreground model.

    return unary: (height*width) * 2 matrix.
    """
    sx, sy = imsz
    mask = np.zeros(imsz, dtype=np.int)

    # background
    mask[:edge_width] = 1
    mask[-edge_width:] = 1
    mask[:,:edge_width] = 1
    mask[:,-edge_width:] = 1

    # foreground
    cx = int(center_block_fraction * sx) + 1
    cy = int(center_block_fraction * sy) + 1
    cx = max(cx, 5)
    cy = max(cy, 5)
    cx_start = sx / 2 - cx / 2
    cy_start = sy / 2 - cy / 2
    mask[cx_start:cx_start + cx, cy_start:cy_start + cy] = 2

    pix_mask = mask.flatten()

    f_background = fmat[pix_mask == 1]
    f_foreground = fmat[pix_mask == 2]

    two_gaussians = TwoGaussiansModel()
    two_gaussians.fit_model(f_background, f_foreground)

    return two_gaussians.make_unary(fmat)

def get_image_dependent_unary(imlist, flist, edge_width, center_block_fraction):
    """
    Compile a list of image dependent unary potentials for a list of images.

    imlist: a list of images
    flist: a list of feature matrices
    edge_width, center_block_fraction: same as those used in
        get_image_dependent_unary_for_single_image.

    return: unary, a list of unary matrices
    """
    u = []
    n_imgs = len(imlist)

    for i in range(n_imgs):
        u.append(get_image_dependent_unary_for_single_image(
            imlist[i].shape[:2], flist[i], edge_width, center_block_fraction))

    return u

def tune_image_dependent_unary(imlist, flist, ground_truth, edge_width_range, 
        center_block_fraction_range, verbose=1):
    """Tune the two parameters edge_width and center_block_fraction for image
    dependent unary model using a labeled dataset.

    imlist: a list of N images
    flist: a list of N feature matrices
    ground_truth: N*M matrix, M is the number of pixels in each image
    edge_width_range: 1-D vector, each element is a possible edge_width
    center_block_fraction_range: 1-D vector, each element is a center_block_fraction
    verbose: print accuracy for each setting if > 0

    return: best_edge_width, best_center_block_fraction
    """
    n_edge_widths = edge_width_range.size
    n_center_block_fractions = center_block_fraction_range.size

    best_acc = 0
    best_edge_width = 0
    best_center_block_fraction = 0

    for i in range(n_edge_widths):
        for j in range(n_center_block_fractions):
            edge_width = edge_width_range[i]
            center_block_fraction = center_block_fraction_range[j]

            u = get_image_dependent_unary(imlist, flist, edge_width, center_block_fraction)

            acc = (gen.unary_predict(u) == ground_truth).mean()
            if verbose > 0:
                print 'edge_width=%g, center_block_fraction=%.2f, acc=%.4f' % (
                        edge_width, center_block_fraction, acc),
                if acc > best_acc:
                    best_acc = acc
                    best_edge_width = edge_width
                    best_center_block_fraction = center_block_fraction
                    print '*'
                else:
                    print ''

    print '-----------------------------------------'
    print 'Best edge_width=%g, center_block_fraction=%g, acc=%.4f' % (
            best_edge_width, best_center_block_fraction, best_acc)

    return best_edge_width, best_center_block_fraction

def tune_combined_unary(original_unary, image_dependent_unary, ground_truth, weight_range):
    """Find the optimal weight to combine an original unary potential with an 
    image dependent unary.

    original_unary: a list of unary potentials
    image_dependent_unary: a list of image dependent unaries
    ground_truth: a matrix of segmentation
    weight_range: a sequence of weight parameters to search

    Return: best_weight, the optimal weight, it will be a tuple 
        [original_unary_weight, image_dependent_unary_weight]
    """

    best_weight = 0
    best_acc = 0

    for weight in weight_range:
        u = un.combine_unary_potentials([original_unary, image_dependent_unary], [1, weight])
        y = gen.unary_predict(u)
        acc = (y == ground_truth).mean()

        print 'image dependent unary weight %.2f, acc=%.4f' % (weight, acc),
        if acc > best_acc:
            best_acc = acc
            best_weight = weight
            print '*'
        else:
            print ''

    # special baseline case
    u = un.combine_unary_potentials([original_unary, image_dependent_unary], [0, 1])
    y = gen.unary_predict(u)
    acc = (y == ground_truth).mean()

    print 'image dependent unary weight  inf, acc=%.4f' % acc

    print '-----------------------------------------------------'
    print 'Best weight %s, acc=%.4f' % (str(best_weight), best_acc)

    return best_weight

def combine_unary(original_unary, image_dependent_unary, weight):
    """Combine an original unary potential with an image dependent unary
    potential.
    
    original_unary, image_dependent_unary: lists of unary potentials
    weight: weight for image_dependent_unary, a tuple [original_unary_weight,
        image_dependent_unary_weight]

    Return: a combined unary potential
    """
    return un.combine_unary_potentials([original_unary, image_dependent_unary],
            [1, weight])

