"""
Manifold regularization for segmentation.

Basic binary segmentation for all images.  All images should be of the same
size.

Yujia Li, 08/2013, 10/2014
"""


import os
import numpy as np
import gnumpy as gnp
import pygco
import cPickle as pickle
import gnn.nn as nn
import imgtools.filter as ft
import imgtools.pairwise as pw
import imgtools.unary as un
import crf.general as gen
import time
import datatools as dt
import crf.cardinality as card
import misc


PRINT_NOTHING   = -1
PRINT_OBJECTIVE = 0
PRINT_LOSS      = 1
PRINT_ALL       = 10

SMALL_CONSTANT = 1e-10

STOPPING_IOU_THRESHOLD = 5e-2
STOPPING_ACC_DIFF_THRESHOLD = 0.05
STOPPING_ACC_STD_THRESHOLD = 4
# STOPPING_ACC_STD_THRESHOLD = 8
STOPPING_TIME_THRESHOLD = 500
STOPPING_TIME_BUDGET = 3600 * 5

ACC_BUFFER_SIZE = 30

#_debug_dd = True
_debug_dd = False
#_debug_show_imgs = True
_debug_show_imgs = False
#_debug_save_figs = True
_debug_save_figs = False

_IMPROVED_COLOR_BEGINS = '\033[42m'
_COLOR_RESET = '\033[0m'

def _colored_txt(txt):
    return _IMPROVED_COLOR_BEGINS + txt + _COLOR_RESET

class Data(object):
    """A general purpose container for data."""
    def __init__(self):
        pass

# a list of pairwise potential generators
pw_gen_list = [
        pw.get_uniform_smoothness_pw, 
        pw.get_boykov_jolly_pw]

class ManifoldRegularizationSeg(object):
    """
    A class that implements manifold regularization for binary segmentation.

    This class keeps a model of the segmentatin system. A model is composed of:
    - model for unary potentials, which is a neural net
    - model for pairwise potentials, which is a list of weights for different
        types of pairwise potentials. The list of pairwise potentials is fixed.
    """
    
    def __init__(self):
        pass

    def load_model(self, model_file):
        """Load a learned model from a file. The model file should contain a
        dictionary with attributes: unary_model, std, avg, pairwise_weights and
        sigma."""
        f = open(model_file)
        d = pickle.load(f)
        f.close()

        self.unary_model = nn.NN()
        self.unary_model.build_net_from_copy(d['unary_model'])
        self.std = d['std']
        self.avg = d['avg']
        self.pairwise_weights = d['pairwise_weights']
        self.boykov_jolly_sigma = d['sigma']

        # if using image dependent unary model
        if 'edge_width' in d:
            self.edge_width = d['edge_width']
            self.center_block_fraction = d['center_block_fraction']
            self.image_dependent_weight = d['image_dependent_weight']

    def write_model(self, model_file):
        """Write the current model to a file."""
        f = open(model_file, 'w')
        if hasattr(self, 'use_image_dependent_unary') and \
                self.use_image_dependent_unary:
            pickle.dump({
                'unary_model' : self.unary_model.make_copy(),
                'std' : self.std,
                'avg' : self.avg,
                'pairwise_weights' : self.pairwise_weights,
                'sigma' : self.boykov_jolly_sigma,
                'edge_width' : self.edge_width,
                'center_block_fraction' : self.center_block_fraction,
                'image_dependent_weight' : self.image_dependent_weight}, f, -1)
        else:
            pickle.dump({
                'unary_model' : self.unary_model.make_copy(),
                'std' : self.std,
                'avg' : self.avg,
                'pairwise_weights' : self.pairwise_weights,
                'sigma' : self.boykov_jolly_sigma}, f, -1)
        f.close()

    def get_model_copy(self):
        """Return a copy of the learned model. It will be an object with four 
        attributes: unary_model, std, avg, pairwise_weights and sigma."""
        model = Data()
        model.unary_model = self.unary_model.make_copy()
        model.std = self.std
        model.avg = self.avg
        model.pairwise_weights = self.pairwise_weights
        model.sigma = self.boykov_jolly_sigma
        
        return model

    def load_model_from_copy(self, model):
        """Load the model from a copy."""
        self.unary_model.build_net_from_copy(model.unary_model)
        self.std = model.std
        self.avg = model.avg
        self.pairwise_weights = model.pairwise_weights
        self.boykov_jolly_sigma = model.sigma

    def load_image_dependent_model(self, model_file):
        """Load image dependent unary model from file."""
        f = open(model_file)
        d = pickle.load(f)
        f.close()

        self.edge_width = d['edge_width']
        self.center_block_fraction = d['center_block_fraction']
        self.image_dependent_weight = d['image_dependent_weight']

    def predict_from_image(self, imgs):
        """
        Make predictions for a set of images using the current model. Note the 
        full pipeline of feature extraction will be run on this set of images
        so this method will be quite expensive to run.

        imgs: a list of N images
        return: y, a N*D matrix, D is the number of pixels in each image
        """
        flist = ft.filter_response(imgs, verbose=0)
        return self.predict_from_flist_and_image(flist, imgs)

    def assemble_pairwise(self, imgs):
        """Assemble pairwise potentials using the current pairwise weights.
        """
        p = []
        n_pw = len(pw_gen_list)
        for k in range(n_pw - 1):
            pw_gen = pw_gen_list[k]
            p.append(pw_gen(imgs))
        p.append(pw_gen_list[n_pw-1](imgs, self.boykov_jolly_sigma))

        return pw.combine_pw(p, self.pairwise_weights)

    def predict_from_flist_and_image(self, flist, imgs):
        """
        Make predictions for a set of images using the current model. The 
        feature list is required so this method does not need to do feature
        extraction any more, and will be faster than predict_from_image.

        return: y, N*D prediction matrix.
        """
        p = self.assemble_pairwise(imgs)

        return self.predict_from_flist_and_pw(flist, p)

    def predict_from_flist_and_pw(self, flist, pairwise):
        """
        Make predictions for a set of images using the current model. Both the
        feature list and the pairwise potentials are required so this method
        does not to do feature extraction and compute pairwise potentials, and
        therefore will be faster than predict_from_flist_and_image.

        return: y, N*D prediction matrix.
        """
        u = un.get_unary(flist, self.unary_model, self.std, self.avg)

        return gen.unary_pairwise_predict(u, pairwise, 1)

    def predict_using_rbm(self, flist, imgs, rbm, rbm_weight, n_iter=-1):
        """
        Make predictions for a set of images using the current model plus a RBM
        prior using the EM algorithm described in our CVPR'13 paper.

        return: y, N*D prediction matrix.
        """
        u = un.get_unary(flist, self.unary_model, self.std, self.avg)
        p = self.assemble_pairwise(imgs)

        return gen.unary_pairwise_rbm_predict(u, p, 1, rbm, rbm_weight, n_iter)

    def load_unary_model(self, unary, std=None, avg=None):
        """Load initial unary model."""
        self.unary_model = unary
        self.std = std
        self.avg = avg

    def load_pairwise_model(self, pairwise_weights, sigma):
        """Load initial pairwise model"""
        self.sigma = sigma
        self.pairwise_weights = pairwise_weights

    def pixel_loss_augmented_inference(self, unary, pairwise, ground_truth):
        """
        Do loss augmented inference for a set of images using the given unary
        potentials, pairwise potentials and ground_truth label matrix.

        unary: a list of unary potentials, each element is a unary matrix for
            one image.
        pairwise: a list of pairwise potentials, each element is a pairwise 
            potential tuple for one image.
        ground_truth: label matrix, each row is one image.

        return: y, same size as ground_truth, the loss augmented inference
            result.
        """
        n_cases, n_pix = ground_truth.shape
        u = unary[:]
        K = u[0].shape[1]

        # This method uses pixel loss, so the loss augmented part only affects
        # unary potentials. Note the unary and pairwise potentials are energy
        # functions, i.e. negative score functions, so we should flip the sign
        # when doing loss-augmented inference.
        all_ones = -np.ones((n_pix, K), dtype=u[0].dtype)
        for i in range(n_cases):
            all_ones[np.arange(n_pix), ground_truth[i]] = 0
            u[i] = u[i] + all_ones  # not +=, to avoid changing the original unary
            all_ones[np.arange(n_pix), ground_truth[i]] = -1

        return gen.unary_pairwise_predict(u, pairwise, 1)


    def load_labeled_data(self, data_file):
        """Load labeled data. The data in the file should be a dictionary with
        attributes 'imgs', 'hogs', 'segs' and 'flist'. flist is the extracted
        pixel feature vectors and is required to avoid redundant computation."""
        f = open(data_file)
        d = pickle.load(f)
        f.close()

        data = Data()
        data.imgs = d['imgs']
        data.hogs = d['hogs']
        data.segs = d['segs']
        data.flist = d['flist']

        return data

    def load_unlabeled_data(self, data_file, count=None):
        """Same as load_labeled_data, except that there is no 'segs'."""
        f = open(data_file)
        d = pickle.load(f)
        f.close()

        if count == None:
            count = d['imgs'].size

        data = Data()
        data.imgs = d['imgs'][:count]
        data.hogs = d['hogs'][:count]
        data.flist = d['flist'][:count]

        return data

    def load_index_file(self, index_file):
        """Load index file. The file should contain a dictionary with one
        attribute `labeled`, which is a 0-1 vector. When the corresponding
        image is labeled, the element will be 1, otherwise 0."""
        f = open(index_file)
        d = pickle.load(f)
        f.close()

        return d['labeled']

    def init_training(self, cfg):
        """Initialize training by loading data and pretrained models, as well
        as initializing necessary parameters."""

        print 'Loading training data from %s ...' % cfg.train_data_file
        self.train_data = self.load_labeled_data(cfg.train_data_file)

        print 'Loading labeled index from %s ...' % cfg.train_index_file
        self.labeled_mask = self.load_index_file(cfg.train_index_file)

        if cfg.extra_unlabeled:
            print 'Loading extra unlabeled data from %s ...' % cfg.unlabeled_data_file
            self.extra_unlabeled_data = self.load_unlabeled_data(
                    cfg.unlabeled_data_file, cfg.unlabeled_count)

        if cfg.is_val:
            print 'Loading valiation data from %s ...' % cfg.val_data_file
            self.val_data = self.load_labeled_data(cfg.val_data_file)

        if cfg.is_test:
            print 'Loading test data from %s ...' % cfg.test_data_file
            self.test_data = self.load_labeled_data(cfg.test_data_file)

        print 'Loading initial model from %s ...' % cfg.init_model_file
        self.load_model(cfg.init_model_file)

        self.use_image_dependent_unary = cfg.use_image_dependent_unary
        if cfg.use_image_dependent_unary:
            print 'Using image dependent unary ...'
        else:
            print 'Not using image dependent unary ...'

        print ''

        self.init_training_without_loading_data_and_model(cfg)

    def init_training_without_loading_data_and_model(self, cfg):
        """Set necessary parameters only, without loading data and model. Data
        and model loading should be done before calling this method."""
        self.config = cfg

        if cfg.is_output and (not os.path.exists(cfg.output_dir)):
            os.makedirs(cfg.output_dir)

        if self.use_image_dependent_unary:
            self.build_image_dependent_unary()

        self.build_pairwise_potentials()
        self.split_labeled_unlabeled_data()
        self.build_graph_and_fixed_potentials()

    def build_image_dependent_unary(self):
        """Build image dependent unary potentials for training/validation and
        test sets."""
        self.train_data.imgdep_u = misc.get_image_dependent_unary(
                self.train_data.imgs, self.train_data.flist, 
                self.edge_width, self.center_block_fraction)
        if self.config.extra_unlabeled:
            self.extra_unlabeled_data.imgdep_u = misc.get_image_dependent_unary(
                    self.extra_unlabeled_data.imgs, self.extra_unlabeled_data.flist,
                    self.edge_width, self.center_block_fraction)
        if self.config.is_val:
            self.val_data.imgdep_u = misc.get_image_dependent_unary(
                    self.val_data.imgs, self.val_data.flist, 
                    self.edge_width, self.center_block_fraction)
        if self.config.is_test:
            self.test_data.imgdep_u = misc.get_image_dependent_unary(
                    self.test_data.imgs, self.test_data.flist, 
                    self.edge_width, self.center_block_fraction)

    def build_pairwise_potentials(self):
        """Build pairwise potentials for training, validation and test set
        using specified weight parameters."""

        self.train_data.pairwise = self.assemble_pairwise(self.train_data.imgs)
        if self.config.extra_unlabeled:
            self.extra_unlabeled_data.pairwise = self.assemble_pairwise(self.extra_unlabeled_data.imgs)
        if self.config.is_val:
            self.val_data.pairwise = self.assemble_pairwise(self.val_data.imgs)
        if self.config.is_test:
            self.test_data.pairwise = self.assemble_pairwise(self.test_data.imgs)

    def split_labeled_unlabeled_data(self):
        """Split the labeled and unlabeled data apart."""
        mask = self.labeled_mask
        n_train = mask.size
        full_idx = np.arange(n_train)
        labeled_idx = full_idx[mask > 0]
        unlabeled_idx = full_idx[mask == 0]

        n_labeled = labeled_idx.size
        n_unlabeled = unlabeled_idx.size

        labeled_data = Data()
        labeled_data.imgs = self.train_data.imgs[labeled_idx]
        labeled_data.segs = self.train_data.segs[labeled_idx]
        labeled_data.hogs = self.train_data.hogs[labeled_idx]

        # note flist and pairwise are lists, not numpy arrays, converting them
        # to numpy object arrays will support better indexing, but will also
        # take extra memory, which is expensive.
        labeled_data.flist = []
        labeled_data.pairwise = []

        if self.use_image_dependent_unary:
            labeled_data.imgdep_u = []
        for i in range(n_labeled):
            labeled_data.flist.append(self.train_data.flist[labeled_idx[i]])
            labeled_data.pairwise.append(self.train_data.pairwise[labeled_idx[i]])
            if self.use_image_dependent_unary:
                labeled_data.imgdep_u.append(self.train_data.imgdep_u[labeled_idx[i]])
        
        unlabeled_data = Data()
        unlabeled_data.imgs = self.train_data.imgs[unlabeled_idx]
        unlabeled_data.segs = self.train_data.segs[unlabeled_idx]   # debug only
        unlabeled_data.hogs = self.train_data.hogs[unlabeled_idx]

        unlabeled_data.flist = []
        unlabeled_data.pairwise = []
        if self.use_image_dependent_unary:
            unlabeled_data.imgdep_u = []
        for i in range(n_unlabeled):
            unlabeled_data.flist.append(self.train_data.flist[unlabeled_idx[i]])
            unlabeled_data.pairwise.append(self.train_data.pairwise[unlabeled_idx[i]])
            if self.use_image_dependent_unary:
                unlabeled_data.imgdep_u.append(self.train_data.imgdep_u[unlabeled_idx[i]])

        self.n_unlabeled_with_label = n_unlabeled

        if self.config.extra_unlabeled:
            unlabeled_data.imgs = np.r_[unlabeled_data.imgs, self.extra_unlabeled_data.imgs]
            unlabeled_data.hogs = np.r_[unlabeled_data.hogs, self.extra_unlabeled_data.hogs]
            unlabeled_data.flist.extend(self.extra_unlabeled_data.flist)
            unlabeled_data.pairwise.extend(self.extra_unlabeled_data.pairwise)
            
            if self.use_image_dependent_unary:
                unlabeled_data.imgdep_u.extend(self.extra_unlabeled_data.imgdep_u)

        self.labeled_data = labeled_data
        self.unlabeled_data = unlabeled_data

    def compute_distance_matrix(self, feature_mat):
        """Compute the distance matrix for a set of data points represented
        by the N*D feature matrix, where N is the number of data points and D
        is the number of dimensionality.
        
        Euclidean distance is used for this computation.
        
        Return a N*N distance matrix.
        """

        n_cases = feature_mat.shape[0]
        gfeat = gnp.garray(feature_mat)
        gdist = gnp.zeros((n_cases, n_cases))

        for i in range(n_cases):
            v = gfeat[i]
            gdist[i] = ((gfeat - v)**2).sum(axis=1)

        return gdist.asarray()

    def build_labeled_unlabeled_nnlist(self, dist, n_labeled, n_unlabeled):
        """Build the nearest neighbor lists for each image. The images are 
        indexed in a way that labeled images come first, then unlabeled images,
        so labeled images use indices 0 to n_labeled-1, unlabeled images use 
        indices n_labeled to n_labeled+n_unlabeled-1.
        
        Return two lists: nn_labeled_list and nn_unlabeled_list. Each is itself
        a list of size n_unlabeled. The indices in nn_labeled_list are the
        original indices for labeled data, and indices in nn_unlabeled_list are
        shifted by n_labeled, so that they start from 0.
        """
        nn_labeled_list = []
        nn_unlabeled_list = []

        top_k = self.config.top_k

        # first pass
        for i in range(n_unlabeled):
            labeled_list = []
            unlabeled_list = []
            
            idx = np.argsort(dist[i+n_labeled])
            for j in range(top_k + 1):
                if idx[j] < n_labeled:
                    labeled_list.append(idx[j])
                elif idx[j] != i + n_labeled:
                    unlabeled_list.append(idx[j] - n_labeled)

            nn_labeled_list.append(labeled_list)
            nn_unlabeled_list.append(unlabeled_list)

        # complete the graph
        for i in range(n_labeled):
            idx = np.argsort(dist[i])
            for j in range(top_k + 1):
                if idx[j] >= n_labeled and (i not in nn_labeled_list[idx[j]-n_labeled]):
                    nn_labeled_list[idx[j]-n_labeled].append(i)
        for i in range(n_unlabeled):
            for j in nn_unlabeled_list[i]:
                if i not in nn_unlabeled_list[j]:
                    nn_unlabeled_list[j].append(i)

        print '%d labeled-to-unlabeled link' % sum([len(n) for n in nn_labeled_list])
        print '%d unlabeled-to-unlabeled link' % sum([len(n) for n in nn_unlabeled_list])
        print ''

        return nn_labeled_list, nn_unlabeled_list

    def build_fixed_unary_potential(self, nn_labeled_list, labeled_segs):
        """Build fixed unary potentials using the labeled images and nearest
        neighbor graph.
        
        Return a matrix of size (n_unlabeled*n_pix_per_image) * 2, where 
        n_unlabeled is the number of unlabeled images. n_pix_per_image is the
        number of pixels for each image.
        """
        n_unlabeled = len(nn_labeled_list)
        n_pix_per_image = labeled_segs.shape[1]
        n_unlabeled_pixels = n_unlabeled * n_pix_per_image

        unary = np.zeros((n_unlabeled_pixels, 2))
        unary_img = np.ones((n_pix_per_image, 2))
        img_pix_idx = np.arange(n_pix_per_image)

        for i in range(n_unlabeled):
            i_start = n_pix_per_image * i
            for j in nn_labeled_list[i]:
                unary_img[img_pix_idx, labeled_segs[j]] = 0
                unary[i_start:i_start+n_pix_per_image] += unary_img
                unary_img[img_pix_idx, labeled_segs[j]] = 1

        unary[:,1] -= self.config.foreground_bias

        return unary * self.config.sim_weight / 2

    def build_fixed_pairwise_potential(self, nn_unlabeled_list):
        """Build pairwise potentials for unlabeled data.
        
        Return (edges, edge_weights)
        
        The number of edges in the graph is
            n_image_edges * n_pix_per_image + n_images * n_edges_in_image
        """
        n_unlabeled = len(nn_unlabeled_list)
        n_pix_per_image = self.train_data.segs.shape[1]
        n_image_edges = 0
        for i in range(n_unlabeled):
            n_image_edges += len(nn_unlabeled_list[i])
        # remove double counting
        n_image_edges /= 2

        n_edges_in_image = self.train_data.pairwise[0][1].size
        n_total_edges = n_image_edges * n_pix_per_image + n_unlabeled * n_edges_in_image

        edges = np.zeros((n_total_edges, 2), dtype=np.int)
        edge_weights = np.zeros(n_total_edges, dtype=np.float32)

        # first part, intra image edges
        i_start = 0
        for i in range(n_unlabeled):
            # pixel indices should be adapted to the global graph
            edges[i_start:i_start+n_edges_in_image] = \
                    self.unlabeled_data.pairwise[i][0] + i * n_pix_per_image
            edge_weights[i_start:i_start+n_edges_in_image] = \
                    self.unlabeled_data.pairwise[i][1] * self.config.C_star / n_unlabeled
            i_start += n_edges_in_image

        # second part, inter image edges
        for i in range(n_unlabeled):
            for j in nn_unlabeled_list[i]:
                # j start from 0, no need for offset
                if i < j:
                    edges[i_start:i_start+n_pix_per_image] = \
                            np.c_[np.arange(i*n_pix_per_image, (i+1)*n_pix_per_image),
                                  np.arange(j*n_pix_per_image, (j+1)*n_pix_per_image)]
                    edge_weights[i_start:i_start+n_pix_per_image] = self.config.sim_weight / 2
                    i_start += n_pix_per_image

        # check if the number of edges matches
        assert(i_start == n_total_edges)

        return edges, edge_weights

    def build_graph_and_fixed_potentials(self):
        """Build the nearest neighbor graph for unlabeled images."""
        
        # compute nearest neighbors
        n_labeled = len(self.labeled_data.imgs)
        n_unlabeled = len(self.unlabeled_data.imgs)
        n_total = n_labeled + n_unlabeled
        top_k = self.config.top_k

        dist = self.compute_distance_matrix(
                np.r_[self.labeled_data.hogs, self.unlabeled_data.hogs])

        nn_labeled_list, nn_unlabeled_list = \
                self.build_labeled_unlabeled_nnlist(dist, n_labeled, n_unlabeled)

        self.nn_labeled_list = nn_labeled_list
        self.nn_unlabeled_list = nn_unlabeled_list

        self.unlabeled_fixed_unary = self.build_fixed_unary_potential(
                nn_labeled_list, self.labeled_data.segs)
        self.unlabeled_pairwise_edges, self.unlabeled_pairwise_edge_weights = \
                self.build_fixed_pairwise_potential(nn_unlabeled_list)

    def add_unlabeled_unary(self, unary):
        """Add the unary potentials used for each individual images to the 
        global one."""
        new_unary = dt.list_to_mat(unary)
        self.updated_fixed_unary = self.unlabeled_fixed_unary + new_unary * self.config.C_star / len(unary)

    def dual_decomposition_inference(self, edges, edge_weights, unary, pw_cost, n_gc_iter, 
            card_fn, n_dd_iter, learn_rate, unary_split):
        """Infer segmentation for unlabeled images using both the graph and a
        cardinality potential."""
        N, K = unary.shape

        # right now the cardinality potential only supports binary segmentation
        assert K == 2

        gc_unary = unary * unary_split
        dd_unary = unary * (1 - unary_split)

        if self.last_delta == None:
            delta = np.zeros((N,K), dtype=np.float)
        else:
            delta = self.last_delta

        y_best = None
        min_energy = np.inf
        if _debug_dd:
            y_original = pygco.cut_general_graph(edges, edge_weights, unary, pw_cost, n_gc_iter, 'expansion')

        for n_iter in range(n_dd_iter):
            y_gc = pygco.cut_general_graph(edges, edge_weights, gc_unary + delta, pw_cost, n_gc_iter, 'expansion')
            y_card = card.cardinality_map_binary(dd_unary - delta, card_fn)

            e = self.get_energy_single_image(unary, edges, edge_weights, pw_cost, y_gc) + card_fn(y_gc.sum())
            if _debug_dd:
                print 'Energy in dd iteration #%d: %f, proportion of fg in prediction %.4f' % (
                        n_iter+1, e, 1.0 * y_gc.sum() / y_gc.size),
            if e < min_energy:
                y_best = y_gc
                min_energy = e
                # keep the best delta to initialize the next dd inference
                self.last_delta = delta

                if _debug_dd:
                    print ' *'
            else:
                if _debug_dd:
                    print ''

            if _debug_dd:
                import matplotlib.pyplot as plt
                import vistools
                plt.ion()
                n_imgs = y_original.size / 1024
                plt.figure(1); vistools.bwpatchview(y_original.reshape(n_imgs, 1024), [32,32], 20); plt.title('Initial graph prediction')
                plt.figure(2); vistools.bwpatchview(y_gc.reshape(n_imgs, 1024), [32,32], 20); plt.title('DD graph')
                plt.figure(3); vistools.bwpatchview(y_card.reshape(n_imgs, 1024), [32,32], 20); plt.title('DD cardinality')
                plt.draw()
                t = raw_input('')

            delta_inc = delta * 0
            delta_inc[np.arange(N), y_gc] += 1
            delta_inc[np.arange(N), y_card] -= 1

            delta += delta_inc * learn_rate / (n_iter + 1)

        return y_best

    def infer_unknown_labels(self, n_iter):
        """Infer the segmentation for unlabeled images."""
        max_edge_weight = 1.0 * self.unlabeled_pairwise_edge_weights.max()
        edges = self.unlabeled_pairwise_edges
        edge_weights = self.unlabeled_pairwise_edge_weights / (max_edge_weight / 10)
        unary = self.updated_fixed_unary / (max_edge_weight / 10)
        pw_cost = 1 - np.eye(2)

        #print 'average unary scale %f' % np.abs(unary).mean()

        if self.config.use_cardinality:
            fg_proportion = 1.0 * self.labeled_data.segs.sum() / self.labeled_data.segs.size
            n_fg_unlabeled = int(fg_proportion * self.updated_fixed_unary.shape[0])
            #print '%d unlabeled pixels' % self.updated_fixed_unary.shape[0]
            #print 'FG proportion: %.4f' % fg_proportion

            # define cardinality potential
            def card_fn(x):
                return np.maximum(np.abs(x - n_fg_unlabeled) - \
                        n_fg_unlabeled / self.config.cardinality_sensitivity, 0)**2
                #return np.maximum(np.abs(x - n_fg_unlabeled) - n_fg_unlabeled / 5, 0)
                # return (x - n_fg_unlabeled)**2

            y = self.dual_decomposition_inference(edges, edge_weights, unary, pw_cost, n_iter,
                    card_fn, self.config.dd_iter, self.config.dd_learn_rate, self.config.dd_unary_split)
        else:
            y = pygco.cut_general_graph(edges, edge_weights, unary, pw_cost, n_iter, 'expansion')
        
        n_unlabeled = len(self.unlabeled_data.imgs)
        n_pix_per_image = self.labeled_data.segs.shape[1]

        return y.reshape(n_unlabeled, n_pix_per_image)

    def compute_nn_consistency(self, inferred_y):
        """Compute the label consistency of inferred labels according to the
        neighborhood structure."""

        n_imgs, n_pix = inferred_y.shape

        consistency = np.zeros(n_imgs)
        v = np.zeros(n_pix)

        nn_labeled_list = self.nn_labeled_list
        nn_unlabeled_list = self.nn_unlabeled_list

        for i in range(n_imgs):
            n_neighbors = len(nn_labeled_list[i]) + len(nn_unlabeled_list[i])
            v *= 0
            y = inferred_y[i]

            for j in nn_labeled_list[i]:
                v += (self.labeled_data.segs[j] != y)

            for j in nn_unlabeled_list[i]:
                v += (inferred_y[j] != y)

            consistency[i] = (v / n_neighbors).mean()

        return consistency

    def update_model_parameters(self, labeled_feature_mat, unlabeled_feature_mat, 
            labeled_y, labeled_y_star,unlabeled_y, unlabeled_y_star, unary_layer_config):
        """Update model parameters using labeled and unlabeled data, in mini-batches
        """
        n_labeled_pixels = labeled_y.size
        n_unlabeled_pixels = unlabeled_y.size

        n_total_pixels = n_labeled_pixels + n_unlabeled_pixels
        batch_size = n_total_pixels / self.config.num_minibatches

        n_labeled_per_batch = batch_size * n_labeled_pixels / n_total_pixels
        if n_labeled_per_batch < 1:
            n_labeled_per_batch = 1
        n_unlabeled_per_batch = batch_size - n_labeled_per_batch

        y_vec = np.zeros((n_total_pixels,2))
        y_vec[np.arange(n_labeled_pixels), labeled_y.flatten()] = 1
        y_vec[np.arange(n_unlabeled_pixels) + n_labeled_pixels, unlabeled_y.flatten()] = 1
        y_star_vec = np.zeros((n_total_pixels,2))
        y_star_vec[np.arange(n_labeled_pixels), labeled_y_star.flatten()] = 1
        y_star_vec[np.arange(n_unlabeled_pixels) + n_labeled_pixels, unlabeled_y_star.flatten()] = 1
 
        # random permutation of the pixels
        labeled_idx = np.random.permutation(n_labeled_pixels)
        unlabeled_idx = np.random.permutation(n_unlabeled_pixels)

        for i in range(self.config.num_minibatches):
            i_labeled_start = i * n_labeled_per_batch
            i_unlabeled_start = i * n_unlabeled_per_batch
            if i < self.config.num_minibatches - 1:
                i_labeled_end = (i+1) * n_labeled_per_batch
                i_unlabeled_end = (i+1) * n_unlabeled_per_batch
            else:
                i_labeled_end = n_labeled_pixels
                i_unlabeled_end = n_unlabeled_pixels

            g_labeled_feature_mat = gnp.garray(labeled_feature_mat[labeled_idx[i_labeled_start:i_labeled_end]])
            g_unlabeled_feature_mat = gnp.garray(unlabeled_feature_mat[unlabeled_idx[i_unlabeled_start:i_unlabeled_end]])

            n_labeled_in_batch = i_labeled_end - i_labeled_start
            n_unlabeled_in_batch = i_unlabeled_end - i_unlabeled_start

            weights = np.ones(n_labeled_in_batch + n_unlabeled_in_batch)
            weights[:n_labeled_in_batch] = (n_labeled_in_batch + n_unlabeled_in_batch) * 1.0 / n_labeled_in_batch 
            weights[n_labeled_in_batch:] = (n_labeled_in_batch + n_unlabeled_in_batch) * 1.0 / \
                    (n_unlabeled_in_batch + SMALL_CONSTANT) * self.config.C_star / self.config.C

            y_vec_batch = np.r_[y_vec[labeled_idx[i_labeled_start:i_labeled_end]], 
                    y_vec[n_labeled_pixels+unlabeled_idx[i_unlabeled_start:i_unlabeled_end]]]
            y_star_vec_batch = np.r_[y_star_vec[labeled_idx[i_labeled_start:i_labeled_end]], 
                    y_star_vec[n_labeled_pixels+unlabeled_idx[i_unlabeled_start:i_unlabeled_end]]]

            self.unary_model._forward(gnp.concatenate((g_labeled_feature_mat, g_unlabeled_feature_mat), axis=0))
            self.unary_model.output.set_gradient(y_vec_batch, y_star_vec_batch, weights)
            self.unary_model._backprop(unary_layer_config)

    def check_grad(self):
        """External interface for check_grad."""
        n_imgs = 5

        labeled_y = self.labeled_data.segs[:n_imgs]
        pairwise = self.labeled_data.pairwise[:n_imgs]
        flist = self.labeled_data.flist[:n_imgs]

        cfg = self.config
        unary_layer_config = Data()
        unary_layer_config.learn_rate = cfg.learn_rate * cfg.C
        unary_layer_config.weight_decay = 1.0 / cfg.C
        unary_layer_config.momentum = cfg.init_momentum

        self._check_grad(labeled_y, pairwise, flist, unary_layer_config)

    def _check_grad(self, labeled_y, pairwise, flist, unary_layer_config):
        """Check if the gradient of the model is correct using finite differences."""
        net = self.unary_model
        std = self.std
        avg = self.avg

        n_imgs, n_pix = labeled_y.shape

        unary = un.get_unary(flist, net, std, avg)
        original_predicted_y = gen.unary_pairwise_predict(unary, pairwise, 1)

        eps = 1e-4

        W_copy = net.layer[0].W.copy()
        sx, sy = net.layer[0].W.shape

        grad = np.zeros(net.layer[0].W.shape)
        for i in range(sx):
            for j in range(sy):
                w_copy = net.layer[0].W[i,j]

                net.layer[0].W[i,j] = w_copy + eps
                u = un.get_unary(flist, net, std, avg)
                predicted_y = gen.unary_pairwise_predict(u, pairwise, 1)
                loss_plus = self.get_energy_diff(u, pairwise, labeled_y, predicted_y)

                net.layer[0].W[i,j] = w_copy - eps
                u = un.get_unary(flist, net, std, avg)
                predicted_y = gen.unary_pairwise_predict(u, pairwise, 1)
                loss_minus = self.get_energy_diff(u, pairwise, labeled_y, predicted_y)

                grad[i,j] = (loss_plus - loss_minus) / (2 * eps * n_imgs)

                net.layer[0].W[i,j] = w_copy

        grad = grad.flatten()

        y_vec = np.zeros((labeled_y.size, 2))
        y_star_vec = np.zeros(y_vec.shape)

        y_vec[np.arange(labeled_y.size), labeled_y.flatten()] = 1
        y_star_vec[np.arange(labeled_y.size), original_predicted_y.flatten()] = 1

        net._forward(gnp.garray(dt.list_to_mat(flist) - avg) / std)
        net.output.set_gradient(y_vec, y_star_vec, np.ones(y_vec.shape[0]) / n_imgs)
        net._backprop(unary_layer_config)

        net_grad = net.layer[0].dLdW.asarray().flatten()

        print 'Grad scale: %f' % np.abs(grad).mean()
        print 'Net grad scale: %f' % np.abs(net_grad).mean()
        print 'Grad diff: %f' % np.sqrt(((grad - net_grad)**2).mean())

    def get_energy_single_image(self, unary, edges, edge_weights, pw_cost, y):
        """Compute the energy of a prediction for a single image."""
        N = y.size
        e = unary[np.arange(N), y].sum()
        e += (pw_cost[y[edges[:,0]], y[edges[:,1]]] * edge_weights).sum()

        return e

    def get_energy(self, unary, pairwise, y):
        """Compute the energy of a prediction under the current model.

        unary: the unary potentials, which is a list of N unary matrices, N is
            the number of images
        pairwise: the pairwise potentials, a list of N pairwise potentials
        y: N*D prediction matrix, D is the number of pixels in one image
        """
        unary_energy = 0
        pairwise_energy = 0

        n_images, n_pixels = y.shape
        for i in range(n_images):
            unary_energy += unary[i][np.arange(n_pixels), y[i]].sum()
            pairwise_energy += ((y[i][pairwise[i][0][:,0]] != y[i][pairwise[i][0][:,1]]) * pairwise[i][1]).sum()
        
        return unary_energy + pairwise_energy

    def get_loss(self, ground_truth_y, predicted_y):
        """Compute the loss of a prediction."""
        return (ground_truth_y != predicted_y).sum()

    def get_structured_hinge_loss(self, unary, pairwise, ground_truth_y, predicted_y):
        """Compute the structured hinge loss for pairs of ground-truth y and predicted y."""
        return -self.get_energy(unary, pairwise, predicted_y) \
                + self.get_loss(ground_truth_y, predicted_y) \
                + self.get_energy(unary, pairwise, ground_truth_y)

    def get_energy_diff(self, unary, pairwise, ground_truth_y, predicted_y):
        """Compute the difference of energies of the ground truth y and the predicted y."""
        return self.get_energy(unary, pairwise, ground_truth_y) - self.get_energy(unary, pairwise, predicted_y)

    def train(self):
        """Train the model using the specified configurations."""
        cfg = self.config

        t_start = time.time()

        training_start_time = t_start

        unary_layer_config = Data()
        unary_layer_config.learn_rate = cfg.learn_rate * cfg.C
        unary_layer_config.weight_decay = 1.0 / cfg.C
        unary_layer_config.momentum = cfg.init_momentum

        labeled_y = self.labeled_data.segs
        u = un.get_unary(self.labeled_data.flist, self.unary_model, self.std, self.avg)
        labeled_feature_mat = (dt.list_to_mat(self.labeled_data.flist) - self.avg) / self.std

        unlabeled_u = un.get_unary(self.unlabeled_data.flist, self.unary_model, self.std, self.avg)
        unlabeled_feature_mat = (dt.list_to_mat(self.unlabeled_data.flist) - self.avg) / self.std

        if self.use_image_dependent_unary:
            u = misc.combine_unary(u, self.labeled_data.imgdep_u, self.image_dependent_weight)
            unlabeled_u = misc.combine_unary(unlabeled_u, self.unlabeled_data.imgdep_u, self.image_dependent_weight)

        self.print_dataset_info_and_config()

        train_pred = gen.unary_pairwise_predict(u, self.labeled_data.pairwise, 1)
        train_acc = gen.pixel_accuracy(train_pred, labeled_y)
        print 'epoch %d, train acc %.4f iou %.4f,' % (0, train_acc,
                gen.intersection_over_union(train_pred, labeled_y)),
        best_acc = train_acc

        if cfg.is_val:
            # update validation unary first
            val_u = un.get_unary(self.val_data.flist, self.unary_model, self.std, self.avg)
            if self.use_image_dependent_unary:
                val_u = misc.combine_unary(val_u, self.val_data.imgdep_u, self.image_dependent_weight)
            val_pred = gen.unary_pairwise_predict(val_u, self.val_data.pairwise, 1)
            val_acc = gen.pixel_accuracy(val_pred, self.val_data.segs)
            print _colored_txt('val acc %.4f * iou %.4f,' % (val_acc,
                    gen.intersection_over_union(val_pred, self.val_data.segs))),
            best_acc = val_acc
        if cfg.is_test:
            # update test unary first
            test_u = un.get_unary(self.test_data.flist, self.unary_model, self.std, self.avg)
            if self.use_image_dependent_unary:
                test_u = misc.combine_unary(test_u, self.test_data.imgdep_u, self.image_dependent_weight)
            test_pred = gen.unary_pairwise_predict(test_u, self.test_data.pairwise, 1)
            print 'test acc %.4f iou %.4f,' % (gen.pixel_accuracy(test_pred, self.test_data.segs),
                    gen.intersection_over_union(test_pred, self.test_data.segs))

        best_epoch = 0
        best_model = self.get_model_copy()

        if cfg.C_star == 0:
            unlabeled_y = gen.unary_predict(unlabeled_u)
            unlabeled_y_star = unlabeled_y

        if cfg.intelligent_stopping:
            acc_buffer = best_acc * np.ones((ACC_BUFFER_SIZE))
            buffer_ptr = 1
            first_round = True

        # used for intelligent_stopping
        last_graph_acc = 0
        current_graph_acc = 0

        # used for dual decomposition inference
        self.last_delta = None

        for epoch in range(cfg.num_epochs):
            gnp.free_reuse_cache()

            if epoch >= cfg.switch_epoch:
                unary_layer_config.momentum = cfg.final_momentum

            # do loss-augmented inference on labeled images
            labeled_y_star = self.pixel_loss_augmented_inference(
                    u, self.labeled_data.pairwise, labeled_y)

            if cfg.C_star > 0:
                # do graph-cuts on unlabeled images
                if epoch % cfg.freeze_epochs == 0:
                    #unlabeled_y = self.unlabeled_data.segs
                    self.add_unlabeled_unary(unlabeled_u)

                    if cfg.self_training:
                        unlabeled_y = gen.unary_pairwise_predict(
                                unlabeled_u, self.unlabeled_data.pairwise, 1)
                    else:
                        unlabeled_y = self.infer_unknown_labels(cfg.graph_cut_iter)

                    if self.unlabeled_data.segs.shape[0] > 0:
                        last_graph_acc = current_graph_acc
                        current_graph_acc = (unlabeled_y[:self.n_unlabeled_with_label] == self.unlabeled_data.segs).mean()

                    #current_unlabeled_y = unlabeled_y
                #else:
                #    self.add_unlabeled_unary(unlabeled_u)
                #    current_unlabeled_y = self.infer_unknown_labels(cfg.graph_cut_iter)


                # do MAP or loss augmented inference on unlabeled images
                # CHANGED on 03/14/2014
                # if cfg.u_loss_augmented or cfg.self_training:
                if cfg.u_loss_augmented:
                    unlabeled_y_star = self.pixel_loss_augmented_inference(
                            unlabeled_u, self.unlabeled_data.pairwise, unlabeled_y)
                    old_unlabeled_y_star = unlabeled_y_star
                else:
                    unlabeled_y_star = gen.unary_pairwise_predict(
                            unlabeled_u, self.unlabeled_data.pairwise, 1)

            old_unlabeled_u = unlabeled_u

            n_to_filter = cfg.inconsistent_to_filter
            consistency = self.compute_nn_consistency(unlabeled_y)
            consistency_sorted_idx = np.argsort(consistency)

            unlabeled_flist = []
            for i in range(consistency_sorted_idx.size - n_to_filter):
                unlabeled_flist.append(self.unlabeled_data.flist[consistency_sorted_idx[i]])

            unlabeled_feature_mat = (dt.list_to_mat(unlabeled_flist) - self.avg) / self.std

            if _debug_show_imgs:
                import vistools
                import matplotlib.pyplot as plt
                plt.ion()

                if not os.path.exists('figs'):
                    os.makedirs('figs')

                #import ipdb
                #ipdb.set_trace()
                plt.figure(1); vistools.bwpatchview(unlabeled_y, [32,32], 20); plt.title('Inferred labels')
                if _debug_save_figs:
                    plt.savefig('figs/inferred_%d.png' % epoch, bbox_inches='tight', pad_inches=0)
                plt.figure(2); vistools.bwpatchview(unlabeled_y_star, [32,32], 20); plt.title('Model prediction')
                if _debug_save_figs:
                    plt.savefig('figs/predicted_%d.png'% epoch, bbox_inches='tight', pad_inches=0)
                #plt.figure(3); vistools.bwpatchview(unlabeled_y[consistency_sorted_idx[:1000]], [32,32], 20)
                #plt.figure(4); vistools.bwpatchview(self.unlabeled_data.segs, [32,32], 20)
                plt.figure(4); vistools.listpatchview(self.unlabeled_data.imgs, 20); plt.title('Color images')
                if _debug_save_figs:
                    plt.savefig('figs/image.png', bbox_inches='tight', pad_inches=0)
                plt.show()
                #import ipdb
                #ipdb.set_trace()
                raw_input()


            # update model parameters
            if n_to_filter > 0:
                self.update_model_parameters(labeled_feature_mat, unlabeled_feature_mat, 
                        labeled_y, labeled_y_star,
                        unlabeled_y[consistency_sorted_idx[:-n_to_filter]], 
                        unlabeled_y_star[consistency_sorted_idx[:-n_to_filter]], unary_layer_config)
            else:
                self.update_model_parameters(labeled_feature_mat, unlabeled_feature_mat, 
                        labeled_y, labeled_y_star,
                        unlabeled_y[consistency_sorted_idx], 
                        unlabeled_y_star[consistency_sorted_idx], unary_layer_config)


            if cfg.u_loss_augmented:
                unlabeled_y_star = gen.unary_pairwise_predict(
                        unlabeled_u, self.unlabeled_data.pairwise, 1)

            # update unary potentials
            u = un.get_unary(self.labeled_data.flist, self.unary_model, self.std, self.avg)
            unlabeled_u = un.get_unary(self.unlabeled_data.flist, self.unary_model, self.std, self.avg)
            if self.use_image_dependent_unary:
                u = misc.combine_unary(u, self.labeled_data.imgdep_u, self.image_dependent_weight)
                unlabeled_u = misc.combine_unary(unlabeled_u, self.unlabeled_data.imgdep_u, self.image_dependent_weight)

            if cfg.is_output and (epoch + 1) % cfg.epoch_to_save:
                self.write_model('%s/%s' % (cfg.output_dir, 'm%d.pdata' % (epoch+1)))

            if (epoch + 1) % cfg.epoch_to_display == 0:
                train_pred = gen.unary_pairwise_predict(u, self.labeled_data.pairwise, 1)
                train_acc = gen.pixel_accuracy(train_pred, labeled_y)
                train_iou = gen.intersection_over_union(train_pred, labeled_y)
                print 'epoch %d, train acc %.4f iou %.4f,' % (epoch+1, train_acc, train_iou),

                if cfg.verbose > PRINT_OBJECTIVE:
                    print 'obj %.2f,' % self.get_structured_hinge_loss(
                            u, self.labeled_data.pairwise, labeled_y, train_pred),

                if cfg.verbose > PRINT_LOSS:
                    print 'loss %.2f,' % self.get_loss(labeled_y, train_pred),

                if self.unlabeled_data.segs.shape[0] > 0:
                    #print 'graph acc %.4f, iou %.4f, cg acc %.4f iou %.4f, map acc %.4f iou %.4f,' % (
                    print 'graph acc %.4f, iou %.4f, map acc %.4f iou %.4f,' % (
                            (unlabeled_y[:self.n_unlabeled_with_label] == self.unlabeled_data.segs).mean(),
                            gen.intersection_over_union(unlabeled_y[:self.n_unlabeled_with_label], self.unlabeled_data.segs),
                            #(current_unlabeled_y[:self.n_unlabeled_with_label] == self.unlabeled_data.segs).mean(),
                            #gen.intersection_over_union(current_unlabeled_y[:self.n_unlabeled_with_label], self.unlabeled_data.segs),
                            (unlabeled_y_star[:self.n_unlabeled_with_label] == self.unlabeled_data.segs).mean(),
                            gen.intersection_over_union(unlabeled_y_star[:self.n_unlabeled_with_label], self.unlabeled_data.segs)),


                    if cfg.verbose > PRINT_OBJECTIVE:
                        # CHANGED 03/14/2014
                        # if cfg.u_loss_augmented or cfg.self_training:
                        if cfg.u_loss_augmented and cfg.self_training:
                            print 'obj %.2f' % self.get_structured_hinge_loss(
                                    old_unlabeled_u[:self.n_unlabeled_with_label], 
                                    self.unlabeled_data.pairwise[:self.n_unlabeled_with_label],
                                    unlabeled_y[:self.n_unlabeled_with_label], 
                                    old_unlabeled_y_star[:self.n_unlabeled_with_label]),
                        else:
                            print 'obj %.2f' % self.get_energy_diff(
                                    old_unlabeled_u[:self.n_unlabeled_with_label], 
                                    self.unlabeled_data.pairwise[:self.n_unlabeled_with_label],
                                    unlabeled_y[:self.n_unlabeled_with_label], 
                                    unlabeled_y_star[:self.n_unlabeled_with_label]),

                        print 'wscale %.4f,' % self.unary_model.output.W.abs().mean(),

                    if cfg.verbose > PRINT_LOSS:
                        print 'loss %.2f,' % self.get_loss(
                                self.unlabeled_data.segs, unlabeled_y_star[:self.n_unlabeled_with_label]),

                if cfg.is_val:
                    # update validation unary first
                    val_u = un.get_unary(self.val_data.flist, self.unary_model, self.std, self.avg)
                    if self.use_image_dependent_unary:
                        val_u = misc.combine_unary(val_u, self.val_data.imgdep_u, self.image_dependent_weight)
                    val_pred = gen.unary_pairwise_predict(val_u, self.val_data.pairwise, 1)
                    val_acc = gen.pixel_accuracy(val_pred, self.val_data.segs)
                    val_iou = gen.intersection_over_union(val_pred, self.val_data.segs)

                    if val_acc > best_acc:
                        best_acc = val_acc
                        best_epoch = epoch+1
                        best_model = self.get_model_copy()
                        print _colored_txt('val acc %.4f * iou %.4f,' % (val_acc, val_iou)),
                    else:
                        print 'val acc %.4f   iou %.4f,' % (val_acc, val_iou),
                else:
                    if train_acc > best_acc:
                        best_acc = train_acc
                        best_epoch = epoch+1
                        best_model = self.get_model_copy()

                if cfg.is_test:
                    # update test unary first
                    test_u = un.get_unary(self.test_data.flist, self.unary_model, self.std, self.avg)
                    if self.use_image_dependent_unary:
                        test_u = misc.combine_unary(test_u, self.test_data.imgdep_u, self.image_dependent_weight)
                    test_pred = gen.unary_pairwise_predict(test_u, self.test_data.pairwise, 1)
                    print 'test acc %.4f iou %.4f,' % (gen.pixel_accuracy(test_pred, self.test_data.segs),
                            gen.intersection_over_union(test_pred, self.test_data.segs)),

                self.unary_model.display_winc()
                print 'time %.2f' % (time.time() - t_start)

                if cfg.intelligent_stopping:
                    if cfg.is_val:
                        # if IOU goes to 0, stop
                        if val_iou < STOPPING_IOU_THRESHOLD:
                            print '[Stopping] IOU too low: val iou %.4f' % val_iou
                            break
                        # if val acc dropped too much, stop
                        if val_acc < best_acc - STOPPING_ACC_DIFF_THRESHOLD:
                            print '[Stopping] val acc dropped too much: val acc %.4f' % val_acc
                            break

                        # print '[Debug] current_graph_acc %.4f, last_graph_acc %.4f, current < last: %s' % (
                        #         current_graph_acc, last_graph_acc, str(current_graph_acc < last_graph_acc))

                        # do not stop when training is still making progress
                        if self.unlabeled_data.segs.shape[0] == 0 or current_graph_acc < last_graph_acc:

                            acc_buffer[buffer_ptr] = val_acc
                            buffer_ptr = (buffer_ptr + 1) % ACC_BUFFER_SIZE

                            if buffer_ptr == 0:
                                first_round = False

                            # local smoothing - avoid outliers
                            # if not first_round or buffer_ptr >= 3:
                            #    mean_val_acc = (acc_buffer[buffer_ptr-1] + acc_buffer[buffer_ptr-2] + acc_buffer[buffer_ptr-3]) / 3
                            mean_val_acc = val_acc

                            if first_round and buffer_ptr > 5 and best_acc - mean_val_acc > acc_buffer[:buffer_ptr].std() * STOPPING_ACC_STD_THRESHOLD:
                                print '[Stopping] acc steadily decreasing'
                                break
                            elif (not first_round) and best_acc - mean_val_acc > acc_buffer.std() * STOPPING_ACC_STD_THRESHOLD:
                                print '[Stopping] acc steadily decreasing'
                                break

                            if time.time() - t_start > STOPPING_TIME_THRESHOLD and time.time() - training_start_time > STOPPING_TIME_BUDGET / 4:
                                print '[Stopping] one iteration takes too much time'
                                break

                            if time.time() - training_start_time > STOPPING_TIME_BUDGET:
                                print '[Stopping] training time budget exceeded, total time %.2f hrs' % ((time.time() - training_start_time) / 3600)
                                break

                t_start = time.time()

        print '----------------------------------------------------------------------------'

        if cfg.is_val:
            print 'Best validation accuracy %.4f at epoch %d' % (best_acc, best_epoch),
        else:
            print 'Best training accuracy %.4f at epoch %d' % (best_acc, best_epoch),

        self.load_model_from_copy(best_model)

        if cfg.is_test:
            test_u = un.get_unary(self.test_data.flist, self.unary_model, self.std, self.avg)
            if self.use_image_dependent_unary:
                test_u = misc.combine_unary(test_u, self.test_data.imgdep_u, self.image_dependent_weight)
            test_pred = gen.unary_pairwise_predict(test_u, self.test_data.pairwise, 1)
            print '  test accuracy %.4f' % (gen.pixel_accuracy(test_pred, self.test_data.segs))
        else:
            print ''

        return best_acc, best_epoch

    def print_dataset_info_and_config(self):
        """Print some basic information about datasets and the parameter settings."""
        print '%d labeled images' % len(self.labeled_data.imgs)
        print '%d unlabeled images' % len(self.unlabeled_data.imgs)

        if self.config.is_val:
            print '%d validation images' % len(self.val_data.imgs)
        if self.config.is_test:
            print '%d test images' % len(self.test_data.imgs)
        print ''
        self.config.display_parameters()

        # minibatch info
        n_pix_per_image = self.labeled_data.segs.shape[1]
        n_labeled_pixels = self.labeled_data.imgs.size * n_pix_per_image
        n_unlabeled_pixels = self.unlabeled_data.imgs.size * n_pix_per_image
        n_total_pixels = n_labeled_pixels + n_unlabeled_pixels
        batch_size = n_total_pixels / self.config.num_minibatches

        n_labeled_per_batch = batch_size * n_labeled_pixels / n_total_pixels
        if n_labeled_per_batch < 1:
            n_labeled_per_batch = 1
        n_unlabeled_per_batch = batch_size - n_labeled_per_batch

        print '%d labeled, %d unlabeled in each batch' % (n_labeled_per_batch, n_unlabeled_per_batch)
        print ''





